local errType = require 'errType'
local utils = require 'utils'
local ipValidator = require 'ipAddrValidator';
local Network_GetState = require 'methods/Network_GetState'

local Network_SetConfig = {}

local function checkIPsWithMask(addrs)
    for _, addr in ipairs(addrs) do
        local res = ipValidator.getIPType(addr)
        if res == ipValidator.Type.ERROR or res == ipValidator.Type.STRING then
            return 'Argument \'' .. tostring(addr) .. '\' is not valid IPv4 or IPv6 address.'
        end
        if res == ipValidator.Type.IPv4 or res == ipValidator.Type.IPv6 then
            return 'Argument \'' .. tostring(addr) .. '\' has no mask specified.'
        end
    end
end

local function checkIPsWithoutMask(addrs)
    for _, addr in ipairs(addrs) do
        local res = ipValidator.getIPType(addr)
        if res == ipValidator.Type.ERROR or res == ipValidator.Type.STRING then
            return 'Argument \'' ..
            tostring(addr) .. '\' is not valid IPv4 or IPv6 address.'
        end
        if res == ipValidator.Type.IPv4_withMask or res == ipValidator.Type.IPv6_withMask then
            return 'Argument \'' .. tostring(addr) ..
            '\' must not have mask specified.'
        end
    end
end

local function bool2ConfigValue(val)
    if val then
        return 'yes'
    else
        return 'no'
    end
end

local function restartNetworkService()
    local _, err = utils.shell_posix('sudo', nil, 'systemctl', 'restart', 'systemd-networkd')
    if err ~= nil then
        return false
    end
    return true
end

local function flushInterface(iface)
    local _, err = utils.shell_posix('sudo', nil, 'ip', 'addr', 'flush', 'dev', iface)
    if err ~= nil then
        return false
    end
    return true
end

function Network_SetConfig.invoke(args)
    local interfaceName, hwaddr, staticAddresses, dnsServers, gateways,
        ipv6AcceptRA, linkLocalAddr, ipMasquerade, dhcpClient,
        dhcpServer, dhcpServerPoolOffset, dhcpServerPoolSize

    if args ~= nil then
        if utils.isArray(args) then
            interfaceName = args[1]
            hwaddr = args[2]
            staticAddresses = args[3]
            dnsServers = args[4]
            gateways = args[5]
            ipv6AcceptRA = args[6]
            ipMasquerade = args[7]
            dhcpClient = args[8]
            dhcpServer = args[9]
            dhcpServerPoolOffset = args[10]
            dhcpServerPoolSize = args[11]
            linkLocalAddr = args[12]
        else
            interfaceName = args.interface
            hwaddr = args.hwaddr
            staticAddresses = args.staticAddresses
            dnsServers = args.dnsServers
            gateways = args.gateways
            ipv6AcceptRA = args.ipv6AcceptRA
            ipMasquerade = args.ipMasquerade
            dhcpClient = args.dhcpClient
            dhcpServer = args.dhcpServer
            dhcpServerPoolOffset = args.dhcpServerPoolOffset
            dhcpServerPoolSize = args.dhcpServerPoolSize
            linkLocalAddr = args.linkLocalAddr
        end
    end

    if interfaceName ~= nil and type(interfaceName) ~= 'string' then
        return nil, errType.codes.InvalidParams, {message = 'Failed to read parameter \'interface\'.'}
    end
    if hwaddr ~= nil and type(hwaddr) ~= 'string' then
        return nil, errType.codes.InvalidParams, {message = 'Failed to read parameter \'hwaddr\'.'}
    end
    if staticAddresses ~= nil and utils.isArray(staticAddresses) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'staticAddresses\' must be an array.'}
    end
    if dnsServers ~= nil and utils.isArray(dnsServers) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'dnsServers\' must be an array.'}
    end
    if gateways ~= nil and utils.isArray(gateways) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'gateways\' must be an array.'}
    end
    if ipv6AcceptRA ~= nil and utils.isBool(ipv6AcceptRA) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'ipv6AcceptRA\' must have a boolean value.'}
    end
    if ipMasquerade ~= nil and utils.isBool(ipMasquerade) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'ipMasquerade\' must have a boolean value.'}
    end
    if linkLocalAddr ~= nil then
        if type(linkLocalAddr) ~= 'string' then
            return nil, errType.codes.InvalidParams, {message = 'Parameter \'linkLocalAddr\' must be a string.'}
        end
        if linkLocalAddr ~= 'yes' and linkLocalAddr ~= 'no' and linkLocalAddr ~= 'ipv4' and linkLocalAddr ~= 'ipv6' then
            return nil, errType.codes.InvalidParams, {message = 'Parameter \'linkLocalAddr\' is not valid.'}
        end
    end
    if dhcpClient ~= nil and utils.isBool(dhcpClient) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'dhcpClient\' must have boolean value.'}
    end
    if dhcpServer ~= nil and utils.isBool(dhcpServer) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'dhcpServer\' must have boolean value.'}
    end

    if dhcpClient and dhcpServer then
        return nil, errType.codes.InvalidParams,
        {message = 'Parameters \'dhcpServer\' and \'dhcpClient\' cannot be set at the same time.'}
    end

    if dhcpServerPoolOffset ~= nil and (utils.isInt(dhcpServerPoolOffset) == false or dhcpServerPoolOffset < 0) then
        return nil, errType.codes.InvalidParams,
        {message = 'Parameter \'dhcpServerPoolOffset\' must be positive integer value.'}
    end
    if dhcpServerPoolOffset ~= nil and dhcpServer ~= true then
        return nil, errType.codes.InvalidParams,
        {message = 'Parameter \'dhcpServerPoolOffset\' must be set only if \'dhcpServer\' is set.'}
    end
    if dhcpServerPoolSize ~= nil and (utils.isInt(dhcpServerPoolSize) == false or dhcpServerPoolSize < 0) then
        return nil, errType.codes.InvalidParams,
        {message = 'Parameter \'dhcpServerPoolSize\' must be positive integer value.'}
    end
    if dhcpServerPoolSize ~= nil and dhcpServer ~= true then
        return nil, errType.codes.InvalidParams,
        {message = 'Parameter \'dhcpServerPoolSize\' must be set only if \'dhcpServer\' is set.'}
    end

    if staticAddresses then
        local res = checkIPsWithMask(staticAddresses)
        if res then
            return nil, errType.codes.InvalidParams, {message = res}
        end
    end

    if dnsServers then
        local res = checkIPsWithoutMask(dnsServers)
        if res then
            return nil, errType.codes.InvalidParams, {message = res}
        end
    end

    if gateways then
        local res = checkIPsWithoutMask(gateways)
        if res then
            return nil, errType.codes.InvalidParams, {message = res}
        end
    end

    if interfaceName == nil and hwaddr == nil then
        return nil, errType.codes.InvalidParams,
        {message = 'Paramter \'interface\' or parameter \'hwaddr\' must be specified.'}
    end
    if interfaceName ~= nil and hwaddr ~= nil then
        return nil, errType.codes.InvalidParams,
        {message = 'Only one paramter from [\'interface\', \'hwaddr\'] must be specified.'}
    end

    local interfaces = Network_GetState.invoke()
    if interfaces == nil then
        return nil, errType.codes.InternalError, {message = 'Failed to query interfaces.'}
    end

    local interface
    for _, iface in ipairs(interfaces) do
        if interfaceName then
            if iface.name == interfaceName then
                interface = iface
            end
        else
            if iface.hwaddr == hwaddr then
                interface = iface
            end
        end
    end
    if interface == nil then
        if interfaceName == nil then
            return nil, errType.codes.InvalidParams,
            {message = 'Interface with MAC address \'' .. hwaddr .. '\' not found.'}
        else
            return nil, errType.codes.InvalidParams,
            {message = 'Interface \'' .. interfaceName .. '\' not found.'}
        end
    end

    local conf = ''

    conf = conf .. '[Match]\n'
    if interfaceName ~= nil then
        conf = conf .. 'Name=' .. interfaceName .. '\n'
    else
        conf = conf .. 'MACAddress=' .. hwaddr .. '\n'
    end
    conf = conf .. '\n[Network]\n'

    if staticAddresses then
        for _, addr in ipairs(staticAddresses) do
            conf = conf .. 'Address=' .. addr .. '\n'
        end
    end
    if dnsServers then
        for _, dns in ipairs(dnsServers) do
            conf = conf .. 'DNS=' .. dns .. '\n'
        end
    end
    if gateways then
        for _, gw in ipairs(gateways) do
            conf = conf .. 'Gateway=' .. gw .. '\n'
        end
    end
    if ipv6AcceptRA ~= nil then
        conf = conf .. 'IPv6AcceptRA=' .. bool2ConfigValue(ipv6AcceptRA) .. '\n'
    end
    if ipMasquerade ~= nil then
        conf = conf .. 'IPMasquerade=' .. bool2ConfigValue(ipMasquerade) .. '\n'
    end
    if linkLocalAddr ~= nil then
        conf = conf .. 'LinkLocalAddressing=' .. linkLocalAddr .. '\n'
    end
    if dhcpClient ~= nil then
        if dhcpClient then
            conf = conf .. 'DHCP=yes\n'
        else
            conf = conf .. 'DHCP=no\n'
        end
    end
    if dhcpServer ~= nil then
        conf = conf .. 'DHCPServer=' .. bool2ConfigValue(dhcpServer) .. '\n'
        if dhcpServerPoolSize or dhcpServerPoolOffset then
            conf = conf .. '\n[DHCPServer]\n'
            if dhcpServerPoolOffset then
                conf = conf .. 'PoolOffset=' .. dhcpServerPoolOffset .. '\n'
            end
            if dhcpServerPoolSize then
                conf = conf .. 'PoolSize=' .. dhcpServerPoolSize .. '\n'
            end
        end
    end

    local fileName
    if interfaceName then
        fileName = interfaceName .. '.network'
    else
        fileName = Network_GetState.hwaddr2name(hwaddr) .. '.network'
    end


    local _, err, errData = utils.shell_posix('sudo', conf, 'dd', 'of=/etc/systemd/network/' .. fileName)
    if err ~= nil and errData.exitcode ~= 0 then
        return nil, errType.codes.ServerError, {message = 'Failed to write to file \'/etc/systemd/network/' .. fileName .. '\'.'}
    end

    local oldConfig
    if interfaceName then
        oldConfig = '/etc/systemd/network/' .. Network_GetState.hwaddr2name(interface.hwaddr) .. '.network'
    else
        oldConfig = '/etc/systemd/network/' .. interface.name .. '.network'
    end
    if utils.fileExists(oldConfig) then
        _, err = utils.shell_posix('sudo', nil, 'rm', oldConfig)
        if err ~= nil then
            return nil, errType.codes.ServerError, {message = 'Failed to remove old configuration file ' .. oldConfig .. '.'}
        end
    end

    flushInterface(interface.name)

    res = restartNetworkService()
    if res == false then
        return nil, errType.codes.InternalError, {message = 'Failed to restart network service.'}
    end
end

return Network_SetConfig
