local cjson = require 'cjson.safe'
local ngx = require "ngx"

local utils = require 'utils'
local errType = require 'errType'

local echo_method = require 'echo'
local Disks_GetState_method = require 'Disks_GetState'
local Disks_Unmount_method = require 'Disks_Unmount'
local RamUsage_GetState_method = require 'RamUsage_GetState'
local TimeDate_GetState_method = require 'TimeDate_GetState'
local TimeDate_GetUptime_method = require 'TimeDate_GetUptime'
local TimeDate_SetConfig_method = require 'TimeDate_SetConfig'
local LoadAverage_GetState_method = require 'LoadAverage_GetState'
local Hostname_GetState_method = require 'Hostname_GetState'
local Hostname_SetConfig_method = require 'Hostname_SetConfig'
local Cpu_GetState_method = require 'Cpu_GetState'
local RexCore_SetConfig_method = require 'RexCore_SetConfig'
local RexCore_GetState_method = require 'RexCore_GetState'
local System_Shutdown_method = require 'System_Shutdown'
local FileSystem_List_method = require 'FileSystem_List'
local FileSystem_GetFileContent_method = require 'FileSystem_GetFileContent'
local FileSystem_SetFileContent_method = require 'FileSystem_SetFileContent'
local FileSystem_Move_method = require 'FileSystem_Move'
local FileSystem_Copy_method = require 'FileSystem_Copy'
local FileSystem_CreateDirectory_method = require 'FileSystem_CreateDirectory'
local FileSystem_CreateFile_method = require 'FileSystem_CreateFile'
local FileSystem_Remove_method = require 'FileSystem_Remove'
local Certificate_Read_method = require 'Certificate_Read'
local Certificate_Write_method = require 'Certificate_Write'
local Network_GetState_method = require 'Network_GetState'
local Network_SetConfig_method = require 'Network_SetConfig'
local Network_RemoveConfig_method = require 'Network_RemoveConfig'
local Network_SetWifiConfig_method = require 'Network_SetWifiConfig'
local Network_ScanWifi_method = require 'Network_ScanWifi'
local Syslog_GetState_method = require 'Syslog_GetState'
local OSInfo_GetState_method = require 'OSInfo_GetState'
local RexArc_Scan_method = require 'RexArc_Scan'
local RexArc_Read_method = require 'RexArc_Read'
local ArcView_ArchivesConfigCreate_method = require 'methods/ArcView_ArchivesConfigCreate'
local ArcView_CheckInitialized_method = require 'methods/ArcView_CheckInitialized'
local ArcView_CheckInstalled_method = require 'methods/ArcView_CheckInstalled'
local ArcView_DbConfigCreate_method = require 'methods/ArcView_DbConfigCreate'
local ArcView_DbPSqlConnectionCheck_method = require 'methods/ArcView_DbPSqlConnectionCheck'
local ArcView_DbPSqlCheckInstalled_method = require 'methods/ArcView_DbPSqlCheckInstalled'
local ArcView_DbPSqlSchemaDataModel_method = require 'methods/ArcView_DbPSqlSchemaDataModel'
local ArcView_DbPSqlSchemaInit_method = require 'methods/ArcView_DbPSqlSchemaInit'
local ArcView_DbPSqlUserInit_method = require 'methods/ArcView_DbPSqlUserInit'
local ArcView_Reset_method = require 'methods/ArcView_Reset'
local ArcView_Restart_method = require 'methods/ArcView_Restart'
local RexStudio_CheckInstalled_method = require 'methods/RexStudio_CheckInstalled'
local RexStudio_Restart_method = require 'methods/RexStudio_Restart'

local methods = {
    echo = echo_method,
    Disks_GetState = Disks_GetState_method,
    Disks_Unmount = Disks_Unmount_method,
    RamUsage_GetState = RamUsage_GetState_method,
    TimeDate_GetState = TimeDate_GetState_method,
    TimeDate_GetUptime = TimeDate_GetUptime_method,
    LoadAverage_GetState = LoadAverage_GetState_method,
    TimeDate_SetConfig = TimeDate_SetConfig_method,
    Hostname_GetState = Hostname_GetState_method,
    Hostname_SetConfig = Hostname_SetConfig_method,
    Network_GetState = Network_GetState_method,
    Cpu_GetState = Cpu_GetState_method,
    RexCore_GetState = RexCore_GetState_method,
    RexCore_SetConfig = RexCore_SetConfig_method,
    System_Shutdown = System_Shutdown_method,
    FileSystem_List = FileSystem_List_method,
    FileSystem_GetFileContent = FileSystem_GetFileContent_method,
    FileSystem_SetFileContent = FileSystem_SetFileContent_method,
    FileSystem_Move = FileSystem_Move_method,
    FileSystem_Copy = FileSystem_Copy_method,
    FileSystem_CreateDirectory = FileSystem_CreateDirectory_method,
    FileSystem_CreateFile = FileSystem_CreateFile_method,
    FileSystem_Remove = FileSystem_Remove_method,
    Certificate_Read = Certificate_Read_method,
    Certificate_Write = Certificate_Write_method,
    Network_SetConfig = Network_SetConfig_method,
    Network_RemoveConfig = Network_RemoveConfig_method,
    Network_SetWifiConfig = Network_SetWifiConfig_method,
    Network_ScanWifi = Network_ScanWifi_method,
    Syslog_GetState = Syslog_GetState_method,
    OSInfo_GetState = OSInfo_GetState_method,
    RexArc_Scan = RexArc_Scan_method,
    RexArc_Read = RexArc_Read_method,
    ArcView_ArchivesConfigCreate = ArcView_ArchivesConfigCreate_method,
    ArcView_CheckInitialized = ArcView_CheckInitialized_method,
    ArcView_CheckInstalled = ArcView_CheckInstalled_method,
    ArcView_DbConfigCreate = ArcView_DbConfigCreate_method,
    ArcView_DbPSqlConnectionCheck = ArcView_DbPSqlConnectionCheck_method,
    ArcView_DbPSqlCheckInstalled = ArcView_DbPSqlCheckInstalled_method,
    ArcView_DbPSqlSchemaDataModel = ArcView_DbPSqlSchemaDataModel_method,
    ArcView_DbPSqlSchemaInit = ArcView_DbPSqlSchemaInit_method,
    ArcView_DbPSqlUserInit = ArcView_DbPSqlUserInit_method,
    ArcView_Reset = ArcView_Reset_method,
    ArcView_Restart = ArcView_Restart_method,
    RexStudio_CheckInstalled = RexStudio_CheckInstalled_method,
    RexStudio_Restart = RexStudio_Restart_method
}

local function createErrorResponse(id, errCode)
    local resp = {
        jsonrpc = '2.0',
        error = {
            code = errCode,
            message = errType.getMsgFromCode(errCode)
        },
        id = id
    }

    if id == nil then
        resp.id = cjson.null
    end

    return resp
end

local function createErrorResponseForRequest(id, errCode, errData)
    -- notification
    if id == nil then
        return nil
    end

    local resp = createErrorResponse(id, errCode)
    resp.id = id;

    if errData ~= nil then
        resp.error.data = errData
    end

    return resp
end

local function checkParams(params)
    if params == nil then
        return true
    end

    if type(params) == 'table' then
        return true
    end

    return false
end

local function checkRequest(req)

    if req.method == nil or type(req.method) ~= 'string' then
        return false
    end

    if req.jsonrpc == nil or type(req.jsonrpc) ~= 'string' or req.jsonrpc ~= '2.0' then
        return false
    end

    return true
end

local function handleRequest(req)

    if type(req) ~= 'table' then
        return createErrorResponse(nil, errType.codes.InvalidRequest)
    end

    if checkRequest(req) == false then
        return createErrorResponse(req.id, errType.codes.InvalidRequest)
    end

    if checkParams(req.params) == false then
        return createErrorResponse(req.id, errType.codes.InvalidRequest)
    end

    local methodArg = req.method

    for key, method in pairs(methods) do
        if key == methodArg then
            local res, err, errData = method.invoke(req.params)

            if err ~= nil then
                return createErrorResponseForRequest(req.id, err, errData)
            else
                -- notification
                if req.id == nil then
                    return nil;
                end

                local result = '';
                if res ~= nil then
                    result = res
                end
                return {
                    jsonrpc = '2.0',
                    id = req.id,
                    result = result
                }
            end
        end
    end

    return createErrorResponseForRequest(req.id, errType.codes.MethodNotFound, nil)
end





ngx.req.read_body()

if ngx.req.get_method() ~= 'POST' then
    ngx.exit(ngx.HTTP_NOT_ALLOWED)
end

local buf = ngx.req.get_body_data()

--buf == nil => no body or body was stored in temp file
if buf == nil then
    local fileName = ngx.req.get_body_file()
    local file = io.open(fileName, "rb")
    if not file then
        return nil
    end
    buf = file:read("*a")
    file:close()
end

if buf == nil then
    local resp = createErrorResponse(nil, errType.codes.InvalidRequest)
    ngx.say(cjson.encode(resp))
    ngx.exit(ngx.HTTP_OK)
end

local body = cjson.decode(buf)

if body == nil or type(body) ~= 'table' then
    local resp = createErrorResponse(nil, errType.codes.ParseError)
    ngx.say(cjson.encode(resp))
    ngx.exit(ngx.HTTP_OK)
end

if utils.isArray(body) then

    if #body == 0 then
        local resp = createErrorResponse(nil, errType.codes.InvalidRequest)
        ngx.say(cjson.encode(resp))
        ngx.exit(ngx.HTTP_OK)
    end

    -- MULTITHREADED APPROACH
    -- local threads = {}
    -- for k, req in pairs(body) do
    --     local thread = ngx.thread.spawn(handleRequest, req)
    --     table.insert(threads, thread)
    -- end
    -- local responses = {}
    -- for i = 1, #threads do
    --     local ok, res = ngx.thread.wait(threads[i])
    --     if ok then
    --         if res ~= nil then
    --             table.insert(responses, res)
    --         end
    --     else
    --         local req = body[i]
    --         if type(req) == 'table' then
    --             table.insert(responses, createErrorResponseForRequest(req.id, errType.codes.ServerError, nil))
    --         else
    --             table.insert(responses, createErrorResponse(nil, errType.codes.ServerError))
    --         end
    --     end
    -- end

    local responses = {}

    for _, req in pairs(body) do
        local response = handleRequest(req)
        if response then
            table.insert(responses, response)
        end
    end

    if #responses > 0 then
        ngx.say(cjson.encode(responses))
    end

    ngx.exit(ngx.HTTP_OK)
else
    local response = handleRequest(body)
    if response ~= nil then
        ngx.say(cjson.encode(response))
    end
    ngx.exit(ngx.HTTP_OK)
end
