local errType = require 'errType'
local config = require 'config'
require 'pipe'

local utils = {}

function utils.isArray(t)

    if type(t) ~= 'table' then
        return false
    end

    local count = 0
    for k, _ in pairs(t) do
        if type(k) ~= 'number' then
            return false
        else
            count = count + 1
        end
    end
    --all keys are numerical. now let's see if they are sequential and start with 1
    for i = 1, count do
        if not t[i] and type(t[i]) ~= 'nil' then
            return false
        end
    end
    return true, count
end

function utils.isInt(n)
    return type(n) == 'number' and n == math.floor(n)
end

function utils.isBool(b)
    return b == true or b == false
end

function utils.split(str, inSplitPattern)
    local outResults = {}
    local theStart = 1
    local theSplitStart, theSplitEnd = string.find(str, inSplitPattern, theStart)
    while theSplitStart do
        table.insert(outResults, string.sub(str, theStart, theSplitStart - 1))
        theStart = theSplitEnd + 1
        theSplitStart, theSplitEnd = string.find(str, inSplitPattern, theStart)
    end
    table.insert(outResults, string.sub(str, theStart))
    return outResults
end

function utils.trim(s)
    if s then
        return s:match "^%s*(.-)%s*$"
    end
    return s
end

function utils.unquote(s)
    if s then
        return s:match '^["\']*(.-)["\']*$'
    end
    return s
end

function utils.startsWith(String,Start)
    return string.sub(String,1,string.len(Start))==Start
end

function utils.endsWith(String,End)
    return End=='' or string.sub(String,-string.len(End))==End
end

function utils.generateErrorData(msg, exitcode, stderr, stdout)
    local data = {}
    data.message = msg
    if exitcode ~= nil then
        data.exitcode = exitcode
    end
    if stderr ~= nil and stderr ~= '' then
        data.stderr = stderr
    end
    if stdout ~= nil and stdout ~= '' then
        data.stdout = stdout
    end
    return data
end

function utils.readLinesFromFile(path)
    local f = io.open(path, "r")
    if f ~= nil then
        local content = f:read("*all")
        f:close()

        local lines = utils.split(content, '\n')
        return lines
    end
    return {}
end

function utils.readFromFile(path)
    local f = io.open(path, "r")
    if f ~= nil then
        local content = f:read("*all")
        f:close()
        return content
    end
    return nil
end

function utils.writeToFile(path, content)
    local f = io.open(path, "w")
    if f == nil then
        return false
    end
    f:write(content)
    f:close()
    return true
end

function utils.removeFile(path)
    local res = os.remove(path)
    if res == true then
        return true
    end
    return false
end

function utils.fileExists(path)
    local f = io.open(path, "r")
    if f ~= nil then
        io.close(f) return true
    else
        return false
    end
end

function utils.isFilesystemReadOnly()
    local stdout, err, errData = utils.shell_posix('mount')
    if err ~= nil or stdout == nil then
        return false
    end

    local lines = utils.split(stdout, '\n')
    for _, line in pairs(lines) do
        local setting = line:match('on%s+/%s+.*%((%a%a)')
        if setting then
            if setting == 'ro' then
                return true
            else
                return false
            end
        end
    end
    return false
end

function utils.lockFilesystem()
    local stdout, err, errData = utils.shell_posix('mount', nil, '-o', 'remount,ro', '/')
    if err ~= nil then
        return false
    end
    return true
end

function utils.unlockFilesystem()
    local stdout, err, errData = utils.shell_posix('mount', nil, '-o', 'remount,rw', '/')
    if err ~= nil then
        return false
    end
    return true
end

function utils.reverseTable(tbl)
    for i = 1, math.floor(#tbl / 2) do
        local tmp = tbl[i]
        tbl[i] = tbl[#tbl - i + 1]
        tbl[#tbl - i + 1] = tmp
    end
end

-- obsolete
function utils.shell(argv)
    local cmd = table.concat(argv,' ')

    local file = io.popen(cmd)
    if file == nil then
        return nil, errType.codes.ServerError,
            utils.generateErrorData('Failed to execute a process with shell command.')
    end

    local stdout = file:read('*all')
    local res = {file:close()}

    local execStatus = nil
    local exitType = nil
    local exitCode = nil


    if #res == 3 then
        execStatus = res[1]
        exitType = res[2] -- contains 'exit' or 'signal'
        exitCode = res[3]
    else
        execStatus = res[1]
    end

    if execStatus ~= true then
        return nil, errType.codes.ServerError,
            utils.generateErrorData('Failed to execute a shell command: \'' .. cmd .. '\'', exitCode, stdout)
    end
    return stdout
end

function utils.shell_posix(cmd, stdin, ...)
    local execCmd = cmd
    if cmd == 'sudo' and select("#", ...) > 0 then
        execCmd = select(1, ...)
    end

    if stdin == nil then
        stdin = ''
    end

    local callStatus, exitCode, stdout, stderr = pcall(pipe_single, stdin, cmd, ...)
    if callStatus == false then
        local errMsg = exitCode -- TODO: log error message
        return nil, errType.codes.ServerError,
        utils.generateErrorData('Failed to execute a shell command: \'' .. execCmd .. '\'')
    end

    if exitCode ~= 0 then
        return stdout, errType.codes.ServerError,
            utils.generateErrorData('Shell command result code suggests an error. Command: \'' .. execCmd .. '\'', exitCode, stderr, stdout)
    end
    if stderr ~= '' then
        return stdout, errType.codes.ServerError,
            utils.generateErrorData('Shell command exits with non-empty error output. Command: \'' .. execCmd .. '\'', exitCode, stderr, stdout)
    end

    return stdout
end

function utils.isPathAbsolute(path)
    if string.sub(path, 1, 1) ~= '/' then
        return false, errType.codes.InvalidParams, {message = 'Path must be absolute and start with \'/\'.'}
    end

    -- check path does not contain '..'
    local splits = utils.split(path, '/')
    for _, pathSegment in pairs(splits) do
        if pathSegment == '..' then
            return false, errType.codes.InvalidParams, {message = 'Path must be absolute and must not contain \'..\'.'}
        end
    end
    return true
end

function utils.isPathReadable(path)
    local readablePaths = nil
    if config and config.file_system and config.file_system.readable_path_filters then
        readablePaths = config.file_system.readable_path_filters
    end

    return utils.isPathAccessible(path, readablePaths)
end

function utils.isPathWritable(path)
    local writablePaths = nil
    if config and config.file_system and config.file_system.writable_path_filters then
        writablePaths = config.file_system.writable_path_filters
    end

    return utils.isPathAccessible(path, writablePaths)
end

function utils.isPathAccessible(path, pathList)
    local isAbsolute, err, errData = utils.isPathAbsolute(path)
    if isAbsolute == false then
        return nil, err, errData
    end

    local isArr = utils.isArray(pathList)
    if isArr then
        local accessible = false
        for _, pathFilter in pairs(pathList) do
            if utils.startsWith(path, pathFilter) then
                accessible = true
            end
        end

        if accessible == false then
            return nil, errType.codes.InvalidParams, {message = 'Access denied.'}
        end
    end
    return true
end

return utils
